#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class AbstractStorageInterface(object):
    @property
    def name(self):
        "Name of the storage implementation."

    def validate(self, url):
        """ Validates the given storage locator and raises a
            :exc:`django.core.exceptions.ValidationError` if errors occurred.
        """


class FileStorageInterface(AbstractStorageInterface):
    """ Interface for storages that provide access to files and allow the
        retrieval of those.
    """

    def retrieve(self, url, location, path):
        """ Retrieve a remote file from the storage specified by the given `url`
            and location and store it to the given `path`. Storages that don't
            need to actually retrieve and store files, just need to return a
            path to a local file instead of storing it under `path`.

            :param url: the URL denoting the storage itself
            :param location: the location of the file to retrieve on the storage
            :param path: a local path where the file should be saved under;
                         this is used as a *hint*
            :returns: the actual path where the file was stored; in some cases
                      this can be different than the passed ``path``
        """

    def list_files(self, url, location):
        """ Return a list of retrievable files available on the storage located
            at the specified URL and given location.

            :param url: the URL denoting the storage itself
            :param location: a template to find items on the storage
            :returns: an iterable of the storage contents under the specified
                      ``location``
        """


class ConnectedStorageInterface(AbstractStorageInterface):
    """ Interface for storages that do not store "files" but provide access to
        data in a different fashion.
    """

    def connect(self, url, location):
        """ Return a connection string for a remote dataset residing on a
            storage specified by the given `url` and `location`.

            :param url: the URL denoting the storage itself
            :param location: the location of the file to retrieve on the storage
            :returns: a connection string to open the stream to actually
                      retrieve data
        """


class PackageInterface(object):

    @property
    def name(self):
        "Name of the package implementation."

    def extract(self, package_filename, location, path):
        """ Extract a file specified by the ``location`` from the package to the
            given ``path`` specification.

            :param package_filename: the local filename of the package
            :param location: a location *within* the package to be extracted
            :param path: a local path where the file should be saved under;
                         this is used as a *hint*
            :returns: the actual path where the file was stored; in some cases
                      this can be different than the passed ``path``
        """

    def list_contents(self, package_filename, location_regex=None):
        """ Return a list of item locations under the specified location in the
            given package.

            :param package_filename: the local filename of the package
            :param location_regex: a template to find items within the package
            :returns: an iterable of the package contents under the specified
                      ``location``
        """



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2020 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import sys
import json

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction

from eoxserver.resources.coverages.registration.stac import (
    create_product_type_from_stac_item, register_stac_product
)
from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to import or export EOxServer type schemas, consisting of
        coverage types, product types, collection types, mask types and
        browse types.
    """

    def add_arguments(self, parser):
        register_parser = self.add_subparser(parser, 'register')
        # export_parser = self.add_subparser(parser, 'export')
        types_parser = self.add_subparser(parser, 'types')

        for parser in (register_parser, types_parser):
            parser.add_argument(
                'location', nargs=1,
                help='The location of the STAC Items. Mandatory.'
            )
            parser.add_argument(
                '--in', '-i', dest='stdin', action="store_true", default=False,
                help='Read the STAC Item from stdin instead from a file.'
            )

        register_parser.add_argument(
            '--type', '--product-type', '-t', dest='type_name', default=None,
            help=(
                'The name of the product type to associate the product with. '
                'Optional.'
            )
        )
        register_parser.add_argument(
            "--replace", "-r",
            dest="replace", action="store_true", default=False,
            help=(
                "Optional. If the product with the given identifier already "
                "exists, replace it. Without this flag, this would result in "
                "an error."
            )
        )

        types_parser.add_argument(
            '--type', '--product-type', '-t', dest='type_name', default=None,
            help=(
                'The name of the new product type. Optional.'
            )
        )
        register_parser.add_argument(
            "--ignore-existing", "-i",
            dest="ignore_existing", action="store_true", default=False,
            help=(
                "Optional. Ignore the case when a product type already "
                "existed. Otherwise, an error is raised."
            )
        )

        # import_parser.add_argument(
        #     '--ignore-existing', action="store_true", default=False,
        #     help='Ignore already existing types.'
        # )

        # export_parser.add_argument(
        #     '--out', '-o', default=None,
        #     help=(
        #         'Write the exported schema to the provided file. By default '
        #         'the schema is written to stdout.'
        #     )
        # )

    @transaction.atomic
    def handle(self, subcommand, *args, **kwargs):
        if subcommand == "register":
            self.handle_register(*args, **kwargs)
        elif subcommand == "types":
            self.handle_types(*args, **kwargs)

    def handle_register(self, location, stdin, type_name, replace,
                        *args, **kw):
        if stdin:
            location = '.'
            values = [json.load(sys.stdin)]
        else:
            location = location[0]
            with open(location) as f:
                values = json.load(f)

        product, replaced = register_stac_product(
            location, values, type_name, replace=replace
        )
        self.print_msg(
            "Successfully %s product %s" % (
                'replaced' if replaced else 'registered',
                product.identifier
            )
        )

    def handle_types(self, location, stdin, type_name, ignore_existing,
                     *args, **kwargs):
        if stdin:
            values = json.load(sys.stdin)
        else:
            with open(location[0]) as f:
                values = json.load(f)

        product_type, created = create_product_type_from_stac_item(
            values, type_name, ignore_existing
        )
        if created:
            self.print_msg(
                "Successfully created product type %s" % product_type.name
            )
        else:
            self.print_msg(
                "Product type %s already existed" % product_type.name
            )



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from eoxserver.contrib import gdal
from eoxserver.backends.access import get_vsi_path, gdal_open
from eoxserver.backends.util import resolve_storage
from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.registration import base
from eoxserver.resources.coverages.registration.exceptions import (
    RegistrationError
)


class BrowseRegistrator(base.BaseRegistrator):
    def register(self, product_identifier, location, type_name=None):
        try:
            product = models.Product.objects.get(identifier=product_identifier)
        except models.Product.DoesNotExist:
            raise RegistrationError('No such product %r' % product_identifier)

        browse_type = None
        if type_name:
            browse_type = models.BrowseType.objects.get(
                name=type_name,
                product_type=product.product_type
            )

        browse = models.Browse(
            product=product,
            location=location[-1],
            storage=resolve_storage(location[:-1]),
            browse_type=browse_type
        )

        # Get a VSI handle for the browse to get the size, extent and CRS
        # via GDAL
        ds = gdal_open(browse)
        browse.width = ds.RasterXSize
        browse.height = ds.RasterYSize
        browse.coordinate_reference_system = ds.GetProjection()
        extent = gdal.get_extent(ds)
        browse.min_x, browse.min_y, browse.max_x, browse.max_y = extent

        browse.full_clean()
        browse.save()
        return browse



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import json
import os

from django.conf import settings
from django.utils.module_loading import import_string

from eoxserver.backends.config import DEFAULT_EOXS_STORAGE_AUTH_HANDLERS

STORAGE_AUTH_HANDLERS = None


class BaseStorageAuthHandler(object):
    def __init__(self, url, parameters):
        self.url = url
        self.parameters = parameters

    def get_vsi_env(self):
        raise NotImplementedError


class S3StorageAuthHandler(BaseStorageAuthHandler):
    name = 'S3'

    def get_vsi_env(self):
        return dict(
            AWS_HTTPS=os.environ.get('AWS_HTTPS', 'NO'),
            AWS_S3_ENDPOINT=self.url,
            AWS_VIRTUAL_HOSTING=os.environ.get('AWS_VIRTUAL_HOSTING', 'FALSE'),
            AWS_ACCESS_KEY_ID=self.parameters.get('ACCESS_KEY_ID'),
            AWS_SECRET_ACCESS_KEY=self.parameters.get('SECRET_ACCESS_KEY'),
        )


def _setup_storage_auth_handlers():
    """ Setup the storage handlers. Uses the ``EOXS_STORAGE_AUTH_HANDLERS``
        setting which falls back to the ``DEFAULT_EOXS_STORAGE_AUTH_HANDLERS``
    """
    global STORAGE_AUTH_HANDLERS
    specifiers = getattr(
        settings, 'EOXS_STORAGE_AUTH_HANDLERS',
        DEFAULT_EOXS_STORAGE_AUTH_HANDLERS
    )
    STORAGE_AUTH_HANDLERS = [
        import_string(specifier) for specifier in specifiers
    ]


def get_handlers():
    if STORAGE_AUTH_HANDLERS is None:
        _setup_storage_auth_handlers()

    return STORAGE_AUTH_HANDLERS


def get_handler_by_test(locator):
    """ Test the given locator with the configured storage handlers and return
        the storage auth
    """
    if STORAGE_AUTH_HANDLERS is None:
        _setup_storage_auth_handlers()

    for storage_auth_handler_cls in STORAGE_AUTH_HANDLERS:
        try:
            if storage_auth_handler_cls.test(locator):
                return storage_auth_handler_cls(locator)
        except AttributeError:
            pass


def get_handler_class_by_name(name):
    if STORAGE_AUTH_HANDLERS is None:
        _setup_storage_auth_handlers()

    for storage_auth_handler_cls in STORAGE_AUTH_HANDLERS:
        try:
            if storage_auth_handler_cls.name == name:
                return storage_auth_handler_cls
        except AttributeError:
            pass


def get_handler_class_for_model(storage_auth_model):
    return get_handler_class_by_name(storage_auth_model.storage_auth_type)


def get_handler_for_model(storage_auth_model):
    return get_handler_class_for_model(
        storage_auth_model
    )(storage_auth_model.url, json.loads(storage_auth_model.auth_parameters))



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.services.ows.wcs.parameters import (
    CoverageRenderParams, CoverageDescriptionRenderParams
)
from eoxserver.services.subset import Subsets, Trim


class WCS10CoverageDescriptionRenderParams(CoverageDescriptionRenderParams):
    coverage_ids_key_name = "coverage"

    def __init__(self, coverages):
        super(WCS10CoverageDescriptionRenderParams, self).__init__(
            coverages, "1.0.0"
        )


class WCS10CoverageRenderParams(CoverageRenderParams):
    def __init__(self, coverage, bbox, crs, format, response_crs=None, 
                 width=None, height=None, resx=None, resy=None, 
                 interpolation=None):

        super(WCS10CoverageRenderParams, self).__init__(coverage, "1.0.0")
        self._bbox = list(bbox)
        self._crs = crs
        self._format = format
        self._response_crs = response_crs
        self._width = width
        self._height = height
        self._resx = resx
        self._resy = resy
        self._interpolation = interpolation

    coverage_id_key_name = "coverage"

    bbox          = property(lambda self: self._bbox)
    crs           = property(lambda self: self._crs)
    format        = property(lambda self: self._format)
    response_crs  = property(lambda self: self._response_crs)
    width         = property(lambda self: self._width)
    height        = property(lambda self: self._height)
    resx          = property(lambda self: self._resx)
    resy          = property(lambda self: self._resy)
    interpolation = property(lambda self: self._interpolation)

    @property
    def subsets(self):
        return Subsets((
            Trim("x", self._bbox[0], self._bbox[2]),
            Trim("y", self._bbox[1], self._bbox[3]),
        ), crs=self._crs)


    def __iter__(self):
        for k, v in super(WCS10CoverageRenderParams, self).__iter__():
            yield k, v

        yield ("bbox", ",".join(map(str, self.bbox)))
        yield ("crs", self.crs)
        yield ("format", self.format)
        
        if self.response_crs:
            yield ("response_crs", self.response_crs)
        
        if self.width is not None:
            yield ("width", str(self.width))
        
        if self.height is not None:
            yield ("height", str(self.height))
        
        if self.resx is not None:
            yield ("resx", self.resx)

        if self.resy is not None:
            yield ("resy", self.resy)
        
        if self.interpolation:
            yield ("interpolation", self.interpolation)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from lxml import etree
from eoxserver.contrib import mapserver as ms
from eoxserver.services.mapserver.wcs.base_renderer import BaseRenderer
from eoxserver.services.ows.version import Version
from eoxserver.services.exceptions import NoSuchCoverageException
from eoxserver.services.result import result_set_from_raw_data, ResultBuffer



class CoverageDescriptionMapServerRenderer(BaseRenderer):
    """ A coverage description renderer implementation using mapserver.
    """

    versions = (Version(1, 1), Version(1, 0))

    def supports(self, params):
        return (
            params.version in self.versions and
            all(
                not coverage.grid.is_referenceable
                for coverage in params.coverages
            )
        )

    def render(self, params):
        map_ = self.create_map()

        use_name = (params.version == Version(1, 0))

        for coverage in params.coverages:

            # ReferenceableDatasets are not supported in WCS < 2.0
            if coverage.grid.is_referenceable:
                raise NoSuchCoverageException((coverage.identifier,))

            data_locations = self.arraydata_locations_for_coverage(coverage)
            native_format = self.get_native_format(coverage, data_locations)
            layer = self.layer_for_coverage(
                coverage, native_format, params.version
            )
            map_.insertLayer(layer)

        for outputformat in self.get_all_outputformats(not use_name):
            map_.appendOutputFormat(outputformat)

        request = ms.create_request(params)
        raw_result = ms.dispatch(map_, request)
        
        result = result_set_from_raw_data(raw_result)
        # load XML using lxml
        # find and exclude <metadataLink> nodes if present
        # re-encode

        xml_result = etree.fromstring(result[0].data)

        for elem in xml_result.xpath('//*[local-name() = "metadataLink"]'):
            elem.getparent().remove(elem)

        xml_result_data =  etree.tostring(xml_result, pretty_print=True, encoding='UTF-8', xml_declaration=True)
        
        result[0] = ResultBuffer(xml_result_data, result[0].content_type)
        
        return result



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from os.path import join
from uuid import uuid4

from eoxserver.contrib import vsi, gdal, vrt
from eoxserver.processing.gdal.vrt import create_simple_vrt
from eoxserver.processing.gdal import reftools
from eoxserver.render.coverage.objects import Mosaic
from eoxserver.resources.coverages.dateline import wrap_extent_around_dateline


class MosaicConnector(object):
    """ Connector for single file layers.
    """

    def supports(self, mosaic, data_items):
        return isinstance(mosaic, Mosaic)

    def connect(self, mosaic, data_items, layer, options):
        try:
            nodata_values = [
                field.nil_values[0][0]
                for field in mosaic.range_type
            ]
        except IndexError:
            nodata_values = None

        vrt_path = '/vsimem/%s.vrt' % uuid4().hex
        vrt.gdalbuildvrt(
            vrt_path, [
                coverage.arraydata_locations[0].path
                for coverage in mosaic.coverages
            ],
            nodata=nodata_values
        )
        layer.data = vrt_path

    def disconnect(self, coverage, data_items, layer, options):
        if layer.data:
            vsi.remove(layer.data)



#-------------------------------------------------------------------------------
#
#  WPS Bounding-Box Data type
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import re
from itertools import chain
from eoxserver.core.util.rect import Rect
from .data_types import Double
from .crs import CRSType
from .base import Parameter
from django.utils.six import string_types

# pre-compiled regular expression used to eliminate repeated extra white-spaces
_RE_MULTIWS = re.compile(r"\s+")


class BoundingBox(tuple):
    """ Bounding-box class.

    Constructor parameters:
        bbox    N-dimensional bounding box definition:
                    ((xmin,),(xmax,))
                    ((xmin,ymin),(xmax,ymax))
                    ((xmin,ymin,zmin),(xmax,ymax,zmax))
                or instance of the ``Rect`` class.
        crs     optional CRS identifier (URI)
    """

    def __new__(cls, bbox, crs=None):
        # pylint: disable=unused-argument
        if isinstance(bbox, Rect):
            lower, upper = bbox.offset, bbox.upper
        else:
            lower, upper = bbox
        lower = tuple(float(v) for v in lower)
        upper = tuple(float(v) for v in upper)
        if len(lower) != len(upper):
            raise ValueError(
                "Dimension mismatch! Both the lower and upper "
                "corners must have the same dimension!"
            )
        return tuple.__new__(cls, (lower, upper))

    def __init__(self, bbox, crs=None):
        tuple.__init__(self)
        self._crs = crs if crs is not None else getattr(bbox, "crs", None)

    @property
    def crs(self):
        """ Get the bounding-box CRS. """
        return self._crs

    @property
    def lower(self):
        """ Get the bounding-box lower coordinates. """
        return self[0]

    @property
    def upper(self):
        """ Get the bounding-box upper coordinates. """
        return self[1]

    @property
    def dimension(self):
        """ Get the bounding-box dimension. """
        return len(self[0])

    @property
    def as_rect(self):
        """Cast to a `Rect` object. (Available only for the 2D bounding-box)."""
        if self.dimension != 2:
            raise RuntimeError(
                "Only 2D bounding-box can be cast to a rectangle object!"
            )
        return Rect(self[0][0], self[0][1], None, None, self[1][0], self[1][1])

    def __str__(self):
        crs = ", crs=%s" % (self.crs if self.crs is not None else "")
        return "BoundingBox((%s, %s)%s)" % (self.lower, self.upper, crs)


class BoundingBoxData(Parameter):
    """ Bounding-box parameter class

    Constructor parameters:
        identifier  identifier of the parameter.
        title       optional human-readable name (defaults to identifier).
        abstract    optional human-readable verbose description.
        metadata    optional metadata (title/URL dictionary).
        optional    optional boolean flag indicating whether the input
                    parameter is optional or not.
        default     optional default input value. Presence of the
                    default value sets the parameter optional.
        crss        list of accepted CRSs (Coordinate Reference Systems).
                    The CRSs shall be given in form of the integer EPSG
                    codes. Defaults to WGS84 (EPSG:4326).
        dimension   optional dimension of the bounding box coordinates.
                    Defaults to 2.
        resolve_input_references Set this option to False not to resolve
                    input references. By default the references are
                    resolved (downloaded and parsed) transparently.
                    If set to False the references must be handled
                    by the process.
    """
    dtype = Double
    dtype_crs = CRSType

    def __init__(self, identifier, crss=None, dimension=2, default=None,
                 *args, **kwargs):
        super(BoundingBoxData, self).__init__(identifier, *args, **kwargs)
        self.dimension = int(dimension)
        self.crss = tuple(self.parse_crs(crs) for crs in crss or (4326,))
        self.default = self.parse(default) if default is not None else None
        if self.dimension < 1:
            raise ValueError("Invalid bounding box dimension %s!" % dimension)

    @property
    def default_crs(self):
        """ Get the bounding-box default CRS. """
        return self.crss[0]

    def _encode(self, bbox):
        """ Common low-level encoding method."""
        if self.dimension != bbox.dimension:
            raise ValueError(
                "Invalid dimension %s of the encoded bounding box!" %
                bbox.dimension
            )
        crs = bbox.crs if bbox.crs is not None else self.default_crs
        if crs not in self.crss:
            raise ValueError("Invalid CRS %s of the encoded bounding box!" % crs)

        return (
            (self.dtype.encode(v) for v in bbox.lower),
            (self.dtype.encode(v) for v in bbox.upper),
            (self.encode_crs(crs),)
        )

    def encode_kvp(self, bbox):
        """ Encode KVP bounding box."""
        return ",".join(chain(*self._encode(bbox)))

    def encode_xml(self, bbox):
        """ Encode XML bounding box."""
        lower, upper, crs = self._encode(bbox)
        return (" ".join(lower), " ".join(upper), crs[0])

    def parse(self, raw_bbox):
        """ Parse the input CRS. """
        if isinstance(raw_bbox, BoundingBox):
            bbox = BoundingBox(
                (raw_bbox.lower, raw_bbox.upper),
                raw_bbox.crs if raw_bbox.crs is not None else self.default_crs
            )
        elif isinstance(raw_bbox, string_types):
            items = raw_bbox.split(',')
            dim = len(items)//2
            lower = [self.dtype.parse(item) for item in items[0:dim]]
            upper = [self.dtype.parse(item) for item in items[dim:2*dim]]
            if len(items) > 2*dim:
                crs = self.parse_crs(items[2*dim])
            else:
                crs = self.default_crs
            bbox = BoundingBox((lower, upper), crs)
        else: # assuming XML decoded tuple
            lower = _RE_MULTIWS.sub(",", raw_bbox[0].strip()).split(",")
            upper = _RE_MULTIWS.sub(",", raw_bbox[1].strip()).split(",")
            if len(lower) != len(upper):
                raise ValueError(
                    "Dimension mismatch of the bounding box's"
                    " corner coordinates! %d != %d" % (len(lower), len(upper))
                )
            lower = [self.dtype.parse(item) for item in lower]
            upper = [self.dtype.parse(item) for item in upper]
            if raw_bbox[2] is not None:
                crs = self.parse_crs(raw_bbox[2])
            else:
                crs = self.default_crs
            bbox = BoundingBox((lower, upper), crs)
        if bbox.dimension != self.dimension:
            raise ValueError(
                "Invalid dimension %d of the parsed bounding box!" %
                bbox.dimension
            )
        if bbox.crs not in self.crss:
            raise ValueError(
                "Invalid CRS %r of the parsed bounding box!" % bbox.crs
            )
        return bbox

    @classmethod
    def parse_crs(cls, raw_crs):
        """ Parse the input bounding CRS. """
        return cls.dtype_crs.parse(raw_crs)

    @classmethod
    def encode_crs(cls, crs):
        """ Encode the output bounding CRS. """
        return cls.dtype_crs.encode(crs)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import os
from django.utils.six import string_types


if os.environ.get('READTHEDOCS', None) != 'True':
    try:
        from osgeo.osr import *
    except ImportError:
        from osr import *

    UseExceptions()

    _SpatialReference = SpatialReference


class SpatialReference(object):
    """ Extension to the original SpatialReference class.
    """

    def __init__(self, raw=None, format=None):
        self.sr = sr = _SpatialReference()
        if raw is not None:
            format = format.upper() if format is not None else None
            if format == "WKT" or (
                isinstance(raw, string_types) and (raw.startswith('PROJCS') or raw.startswith('GEOGCS'))
            ):
                sr.ImportFromWkt(raw)
            elif isinstance(raw, int) or format == "EPSG":
                sr.ImportFromEPSG(int(raw))
            elif isinstance(raw, string_types) and raw.startswith('EPSG:'):
                sr.ImportFromEPSG(int(raw.partition(':')[2]))
            else:
                sr.SetFromUserInput(raw)

    def IsSame(self, other):
        if isinstance(other, SpatialReference):
            return self.sr.IsSame(other.sr)
        else:
            return self.sr.IsSame(other)

    @property
    def proj(self):
        return self.sr.ExportToProj4()

    @property
    def wkt(self):
        return self.sr.ExportToWkt()

    @property
    def xml(self):
        return self.sr.ExportToXML()

    @property
    def url(self):
        # TODO: what about other authorities than EPSG?
        return "http://www.opengis.net/def/crs/EPSG/0/%d" % self.srid

    @property
    def srid(self):
        """ Convenience function that tries to get the SRID of the projection.
        """

        if self.sr.IsGeographic():
            cstype = 'GEOGCS'
        else:
            cstype = 'PROJCS'

        srid = self.sr.GetAuthorityCode(cstype)
        return int(srid) if srid is not None else None

    @property
    def swap_axes(self):
        # TODO:
        pass

    def __getattr__(self, name):
        return getattr(self.sr, name)



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from weakref import proxy

from eoxserver.render.coverage.objects import (
    GRID_TYPE_TEMPORAL, GRID_TYPE_ELEVATION
)


class Layer(object):
    """ Abstract layer
    """
    def __init__(self, name, style):
        self._name = name
        self._style = style
        self._map = None

    @property
    def name(self):
        return self._name

    @property
    def style(self):
        return self._style

    @property
    def map(self):
        return self._map

    @map.setter
    def map(self, map_):
        self._map = proxy(map_)


class CoverageLayer(Layer):
    """ Representation of a coverage layer.
    """
    def __init__(self, name, style, coverage, bands, wavelengths, time,
                 elevation, ranges):
        super(CoverageLayer, self).__init__(name, style)
        self._coverage = coverage
        self._bands = bands
        self._wavelengths = wavelengths
        self._time = time
        self._elevation = elevation
        self._ranges = ranges

    @property
    def coverage(self):
        return self._coverage

    @property
    def bands(self):
        return self._bands

    @property
    def wavelengths(self):
        return self._wavelengths

    @property
    def time(self):
        return self._time

    @property
    def elevation(self):
        return self._elevation

    @property
    def ranges(self):
        return self._ranges


class CoveragesLayer(Layer):
    """ Representation of a coverages layer.
    """
    def __init__(self, name, style, coverages, bands, wavelengths, time,
                 elevation, ranges):
        super(CoveragesLayer, self).__init__(name, style)
        self._coverages = coverages
        self._bands = bands
        self._wavelengths = wavelengths
        self._time = time
        self._elevation = elevation
        self._ranges = ranges

    @property
    def coverages(self):
        return self._coverages

    @property
    def bands(self):
        return self._bands

    @property
    def wavelengths(self):
        return self._wavelengths

    @property
    def time(self):
        return self._time

    @property
    def elevation(self):
        return self._elevation

    @property
    def ranges(self):
        return self._ranges


class OutlinedCoveragesLayer(Layer):
    """ Representation of an outlined coverages layer.
    """
    def __init__(self, name, style, coverages, bands, wavelengths, time,
                 elevation, ranges):
        super(OutlinedCoveragesLayer, self).__init__(name, style)
        self._coverages = coverages
        self._bands = bands
        self._wavelengths = wavelengths
        self._time = time
        self._elevation = elevation
        self._ranges = ranges

    @property
    def coverages(self):
        return self._coverages

    @property
    def bands(self):
        return self._bands

    @property
    def wavelengths(self):
        return self._wavelengths

    @property
    def time(self):
        return self._time

    @property
    def elevation(self):
        return self._elevation

    @property
    def ranges(self):
        return self._ranges


class MosaicLayer(Layer):
    def __init__(self, name, style, mosaic, coverages, bands, wavelengths, time,
                 elevation, ranges):
        super(MosaicLayer, self).__init__(name, style)
        self._mosaic = mosaic
        self._coverages = coverages
        self._bands = bands
        self._wavelengths = wavelengths
        self._time = time
        self._elevation = elevation
        self._ranges = ranges

    @property
    def mosaic(self):
        return self._mosaic

    @property
    def coverages(self):
        return self._coverages

    @property
    def bands(self):
        return self._bands

    @property
    def wavelengths(self):
        return self._wavelengths

    @property
    def time(self):
        return self._time

    @property
    def elevation(self):
        return self._elevation

    @property
    def ranges(self):
        return self._ranges


class BrowseLayer(Layer):
    """ Representation of a browse layer.
    """
    def __init__(self, name, style, browses, ranges=None):
        super(BrowseLayer, self).__init__(name, style)
        self._browses = browses
        self._ranges = ranges

    @property
    def browses(self):
        return self._browses

    @property
    def ranges(self):
        return self._ranges


class OutlinedBrowseLayer(Layer):
    """ Representation of a browse layer.
    """
    def __init__(self, name, style, browses, ranges=None):
        super(OutlinedBrowseLayer, self).__init__(name, style)
        self._browses = browses
        self._ranges = ranges

    @property
    def browses(self):
        return self._browses

    @property
    def ranges(self):
        return self._ranges


class MaskLayer(Layer):
    """ Representation of a mask layer.
    """
    def __init__(self, name, style, masks):
        super(MaskLayer, self).__init__(name, style)
        self._masks = masks

    @property
    def masks(self):
        return self._masks


class MaskedBrowseLayer(Layer):
    """ Representation of a layer.
    """
    def __init__(self, name, style, masked_browses):
        super(MaskedBrowseLayer, self).__init__(name, style)
        self._masked_browses = masked_browses

    @property
    def masked_browses(self):
        return self._masked_browses


class OutlinesLayer(Layer):
    """ Representation of a layer.
    """
    def __init__(self, name, style, fill, footprints, masks=None):
        super(OutlinesLayer, self).__init__(name, style)
        self._footprints = footprints
        self._masks = masks
        self._fill = fill

    @property
    def footprints(self):
        return self._footprints

    @property
    def masks(self):
        return self._masks

    @property
    def fill(self):
        return self._fill


class Map(object):
    """ Abstract interpretation of a map to be drawn.
    """
    def __init__(self, layers, width, height, format, bbox, crs, bgcolor=None,
                 transparent=True, time=None, elevation=None):
        self._layers = layers
        self._width = int(width)
        self._height = int(height)
        self._format = format
        self._bbox = bbox
        self._crs = crs
        self._bgcolor = bgcolor
        self._transparent = transparent
        self._time = time
        self._elevation = elevation

        for layer in layers:
            layer.map = self

    @property
    def layers(self):
        return self._layers

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def format(self):
        return self._format

    @property
    def bbox(self):
        return self._bbox

    @property
    def crs(self):
        return self._crs

    @property
    def bgcolor(self):
        return self._bgcolor

    @property
    def transparent(self):
        return self._transparent

    @property
    def time(self):
        return self._time

    @property
    def elevation(self):
        return self._elevation

    def __repr__(self):
        return (
            'Map: %r '
            'width=%r '
            'height=%r '
            'format=%r '
            'bbox=%r '
            'crs=%r '
            'bgcolor=%r '
            'transparent=%r '
            'time=%r '
            'elevation=%r' % (
                self.layers, self.width, self.height, self.format, self.bbox,
                self.crs, self.bgcolor, self.transparent, self.time,
                self.elevation,
            )
        )


class LayerDescription(object):
    """ Abstract layer description
    """

    is_raster = False

    def __init__(self, name, bbox=None, dimensions=None, queryable=False,
                 styles=None, sub_layers=None, title=None):
        self._name = name
        self._bbox = bbox
        self._dimensions = dimensions if dimensions is not None else {}
        self._queryable = queryable
        self._styles = styles if styles is not None else []
        self._sub_layers = sub_layers if sub_layers is not None else []
        self._title = title

    @property
    def name(self):
        return self._name

    @property
    def bbox(self):
        return self._bbox

    @property
    def dimensions(self):
        return self._dimensions

    @property
    def queryable(self):
        return self._queryable

    @property
    def styles(self):
        return self._styles

    @property
    def sub_layers(self):
        return self._sub_layers

    @property
    def title(self):
        return self._title or self.name

    @classmethod
    def from_coverage(cls, coverage, styles):
        extent = coverage.extent
        grid = coverage.grid

        dimensions = {}
        if GRID_TYPE_ELEVATION in grid.types:
            elevation_dim = grid.types.index(GRID_TYPE_ELEVATION)
            dimensions['elevation'] = {
                'min': extent[elevation_dim],
                'max': extent[len(extent) / 2 + elevation_dim],
                'step': grid.offsets[elevation_dim],
                'default': extent[len(extent) / 2 + elevation_dim],
                'units': 'CRS:'  # TODO: get vertical part of crs
            }

        if GRID_TYPE_TEMPORAL in grid.types:
            temporal_dim = grid.types.index(GRID_TYPE_TEMPORAL)
            dimensions['time'] = {
                'min': extent[temporal_dim],
                'max': extent[len(extent) / 2 + temporal_dim],
                'step': grid.offsets[temporal_dim],
                'default': extent[len(extent) / 2 + temporal_dim],
                'units': 'ISO8601'
            }

        range_type = coverage.range_type
        band_names = [
            field.identifier for field in range_type
        ]
        wavelengths = [
            str(field.wavelength)
            for field in range_type
            if field.wavelength is not None
        ]

        dimensions['bands'] = {'values': band_names}

        if wavelengths:
            dimensions['wavelength'] = {'values': wavelengths}

        return cls(
            coverage.identifier,
            bbox=coverage.footprint.extent if coverage.footprint else None,
            dimensions=dimensions,
            styles=styles
        )

    @property
    def from_browse_type(cls, eo_object, browse_type):
        browse_type



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import re
from datetime import datetime, timedelta

from django.utils.timezone import utc, make_aware, is_aware
#from django.utils.dateparse import parse_datetime, parse_date

try:
    from dateutil.parser import parse
    HAVE_DATEUTIL = True
except ImportError:
    from django.utils.dateparse import parse_datetime, parse_date
    HAVE_DATEUTIL = False


def isoformat(dt):
    """ Formats a datetime object to an ISO string. Timezone naive datetimes are
        are treated as UTC Zulu. UTC Zulu is expressed with the proper "Z"
        ending and not with the "+00:00" offset declaration.

        :param dt: the :class:`datetime.datetime` to encode
        :returns: an encoded string
    """
    if not dt.utcoffset():
        dt = dt.replace(tzinfo=None)
        return dt.isoformat("T") + "Z"
    return dt.isoformat("T")


def parse_iso8601(value, tzinfo=None):
    """ Parses an ISO 8601 date or datetime string to a python date or datetime.
        Raises a `ValueError` if a conversion was not possible. The returned
        datetime is always considered time-zone aware and defaulting to the
        given timezone `tzinfo` or UTC Zulu if none was specified.

        If the optional module :mod:`dateutil` is installed, it is used in
        preference over the :mod:`dateparse <django.utils.dateparse>` functions.

        :param value: the string value to be parsed
        :param tzinfo: an optional tzinfo object that is used when the input
                       string is not timezone aware
        :returns: a :class:`datetime.datetime`
    """

    tzinfo = tzinfo or utc
    parsers = (parse,) if HAVE_DATEUTIL else (parse_datetime, parse_date)
    for parser in parsers:
        try:
            temporal = parser(value)
        except Exception as e:
            raise ValueError(
                "Could not parse '%s' to a temporal value. "
                "Error was: %s" % (value, e)
            )
        if temporal:
            # convert to datetime if necessary
            if not isinstance(temporal, datetime):
                temporal = datetime.combine(temporal, datetime.min.time())

            # use UTC, if the datetime is not already time-zone aware
            if not is_aware(temporal):
                temporal = make_aware(temporal, tzinfo)

            return temporal

    raise ValueError("Could not parse '%s' to a temporal value" % value)


RE_ISO_8601 = re.compile(
    r"^(?P<sign>[+-])?P"
    r"(?:(?P<years>\d+(\.\d+)?)Y)?"
    r"(?:(?P<months>\d+(\.\d+)?)M)?"
    r"(?:(?P<days>\d+(\.\d+)?)D)?"
    r"(T"
    r"(?:(?P<hours>\d+(\.\d+)?)H)?"
    r"(?:(?P<minutes>\d+(\.\d+)?)M)?"
    r"(?:(?P<seconds>\d+(\.\d+)?)S)?"
    r")?$"
)


def parse_duration(value):
    """ Parses an ISO 8601 duration string into a python timedelta object.
        Raises a `ValueError` if a conversion was not possible.
    """

    match = RE_ISO_8601.match(value)
    if not match:
        raise ValueError(
            "Could not parse ISO 8601 duration from '%s'." % value
        )
    match = match.groupdict()

    sign = -1 if "-" == match['sign'] else 1
    days = float(match['days'] or 0)
    days += float(match['months'] or 0) * 30  # ?!
    days += float(match['years'] or 0) * 365  # ?!
    fsec = float(match['seconds'] or 0)
    fsec += float(match['minutes'] or 0) * 60
    fsec += float(match['hours'] or 0) * 3600

    return sign * timedelta(days, fsec)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.core.decoders import (
    kvp, xml, enum, value_range, boolean, InvalidParameterException
)
from eoxserver.core.util.xmltools import NameSpace, NameSpaceMap
from eoxserver.services.ows.wcs.v20.util import ns_wcs

class CompressionNotSupported(InvalidParameterException):
    code = "CompressionNotSupported"


class CompressionInvalid(InvalidParameterException):
    code = "CompressionInvalid"

class JpegQualityInvalid(InvalidParameterException):
    code = "JpegQualityInvalid"

class PredictorInvalid(InvalidParameterException):
    code = "PredictorInvalid"

class PredictorNotSupported(InvalidParameterException):
    code = "PredictorNotSupported"


class InterleavingInvalid(InvalidParameterException):
    code =  "InterleavingInvalid"

class TilingInvalid(InvalidParameterException):
    code = "TilingInvalid"

def parse_jpeg_quality(value):

    value = int(value)
    if value < 1 or value > 100:
        raise JpegQualityInvalid(
            "geotiff:jpeg_quality should be an integer between 1 and 100",
            "geotiff:jpeg_quality")
    return value

class WCS20GeoTIFFEncodingExtension(object):
    def supports(self, frmt, options):
        # To allow "native" GeoTIFF formats aswell
        if not frmt:
            return True
        return frmt.lower() == "image/tiff"

    def get_decoder(self, request):
        if request.method == "GET":
            return WCS20GeoTIFFEncodingExtensionKVPDecoder(request.GET)
        else:
            return WCS20GeoTIFFEncodingExtensionXMLDecoder(request.body)

    def get_encoding_params(self, request):
        decoder = self.get_decoder(request)

        # perform some dependant value checking
        compression = decoder.compression
        predictor = decoder.predictor
        jpeg_quality = decoder.jpeg_quality
        tiling = decoder.tiling
        tileheight = decoder.tileheight
        tilewidth = decoder.tilewidth

        if predictor and compression not in ("LZW", "Deflate"):
            raise PredictorNotSupported(
                "geotiff:predictor requires compression method 'LZW' or "
                "'Deflate'.", "geotiff:predictor"
            )

        if jpeg_quality is not None:  
            if compression != "JPEG":
                raise CompressionNotSupported(
                    "geotiff:jpeg_quality requires compression method 'JPEG'.",
                    "geotiff:jpeg_quality"
                )
            else :
                parse_jpeg_quality(jpeg_quality)

        if tiling and (tileheight is None or tilewidth is None):
            raise TilingInvalid(
                "geotiff:tiling requires geotiff:tilewidth and "
                "geotiff:tileheight to be set.", "geotiff:tiling"
            )

        return {
            "compression": compression,
            "jpeg_quality": jpeg_quality,
            "predictor": predictor,
            "interleave": decoder.interleave,
            "tiling": tiling,
            "tileheight": tileheight,
            "tilewidth": tilewidth
        }


compression_enum = enum(
    ("None", "PackBits", "Huffman", "LZW", "JPEG", "Deflate"), True, CompressionInvalid
)
predictor_enum = enum(("None", "Horizontal", "FloatingPoint"), True, PredictorInvalid)
interleave_enum = enum(("Pixel", "Band"), True, InterleavingInvalid)

def parse_multiple_16(raw):
    value = int(raw)
    if value < 0:
        raise ValueError("Value must be a positive integer.")
    elif (value % 16) != 0:
        raise ValueError("Value must be a multiple of 16.")
    return value


class WCS20GeoTIFFEncodingExtensionKVPDecoder(kvp.Decoder):
    compression = kvp.Parameter("geotiff:compression", num="?", type=compression_enum)
    jpeg_quality = kvp.Parameter("geotiff:jpeg_quality", num="?", type=int)
    predictor   = kvp.Parameter("geotiff:predictor", num="?", type=predictor_enum)
    interleave  = kvp.Parameter("geotiff:interleave", num="?", type=interleave_enum)
    tiling      = kvp.Parameter("geotiff:tiling", num="?", type=boolean)
    tileheight  = kvp.Parameter("geotiff:tileheight", num="?", type=parse_multiple_16)
    tilewidth   = kvp.Parameter("geotiff:tilewidth", num="?", type=parse_multiple_16)


class WCS20GeoTIFFEncodingExtensionXMLDecoder(xml.Decoder):
    compression = xml.Parameter("wcs:Extension/geotiff:parameters/geotiff:compression/text()", num="?", type=compression_enum, locator="geotiff:compression")
    jpeg_quality = xml.Parameter("wcs:Extension/geotiff:parameters/geotiff:jpeg_quality/text()", num="?", type=int, locator="geotiff:jpeg_quality")
    predictor   = xml.Parameter("wcs:Extension/geotiff:parameters/geotiff:predictor/text()", num="?", type=predictor_enum, locator="geotiff:predictor")
    interleave  = xml.Parameter("wcs:Extension/geotiff:parameters/geotiff:interleave/text()", num="?", type=interleave_enum, locator="geotiff:interleave")
    tiling      = xml.Parameter("wcs:Extension/geotiff:parameters/geotiff:tiling/text()", num="?", type=boolean, locator="geotiff:tiling")
    tileheight  = xml.Parameter("wcs:Extension/geotiff:parameters/geotiff:tileheight/text()", num="?", type=parse_multiple_16, locator="geotiff:tileheight")
    tilewidth   = xml.Parameter("wcs:Extension/geotiff:parameters/geotiff:tilewidth/text()", num="?", type=parse_multiple_16, locator="geotiff:tilewidth")

    namespaces = NameSpaceMap(
        ns_wcs, NameSpace("http://www.opengis.net/gmlcov/geotiff/1.0", "geotiff")
    )



from django.core.management.base import BaseCommand, CommandError

from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import CommandOutputMixIn
from eoxserver.services.ecql import parse, to_filter, get_repr
from eoxserver.services.filters import get_field_mapping_for_model


class Command(CommandOutputMixIn, BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument(
            'cql_filters', nargs='+', help='The search filters'
        )

        parser.add_argument(
            '--collection', '-c', dest='collection_id',
            help='Optional. Only list datasets in this collection.'
        )
        parser.add_argument(
            '--type', '-t', dest='type', default='EOObject',
            help='Optional. Limit datasets to objects of that type.'
        )
        parser.add_argument(
            '--exclude', '-e', dest='exclude',
            action='store_true', default=False,
            help=(
                'Optional. Reverse the lookup: instead of including the matched '
                'datasets in the result, exclude them and include everything '
                'else.'
            )
        )

        parser.add_argument(
            '--show-attributes', '--show', '-s', dest='show_attributes',
            action='store_true', default=False,
            help=(
                'Optional. Display the available attributes for the given '
                'record type.'
            )
        )

    args = '<cql-filter>'

    help = """
        Perform a query of datasets matching the given filters expressed in CQL.
        The dataset IDs will be written to stdout.
    """

    def handle(self, cql_filters, **options):
        self.verbosity = int(options.get('verbosity', 1))

        # get the model class and the field mapping (with choices)
        ModelClass = getattr(models, options.get('type'))
        mapping, mapping_choices = get_field_mapping_for_model(ModelClass)

        # print the available attributes, if requested
        if options.get('show_attributes'):
            print("\n".join(mapping.keys()))
            return

        # filter by collection, if requested
        collection_id = options.get('collection_id')
        if collection_id:
            try:
                collection = models.Collection.objects.get(
                    identifier=collection_id
                )
                qs = ModelClass.objects.filter(collections__in=[collection.pk])
            except models.Collection.DoesNotExist:
                raise CommandError('No such collection %r' % collection_id)
        else:
            qs = ModelClass.objects.all()

        if len(cql_filters) < 1:
            raise CommandError('No CQL filter passed.')

        for cql_filter in cql_filters:
            ast = parse(cql_filter)

            if self.verbosity >= 2:
                self.print_msg(get_repr(ast), 2)

            filters = to_filter(ast, mapping, mapping_choices)
            if not filters:
                raise CommandError('Invalid filter specified')

            if options['exclude']:
                qs = ModelClass.objects.exclude(filters)
            else:
                qs = ModelClass.objects.filter(filters)

            if self.verbosity >= 2:
                self.print_msg(filters, 2)

        qs = qs.values_list('identifier', flat=True)
        print ("\n".join(qs))



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from os.path import join
from uuid import uuid4

from eoxserver.backends.access import get_vsi_path
from eoxserver.contrib import vsi, gdal
from eoxserver.processing.gdal.vrt import create_simple_vrt
from eoxserver.processing.gdal import reftools
from eoxserver.resources.coverages.dateline import wrap_extent_around_dateline


class SimpleConnector(object):
    """ Connector for single file layers.
    """

    def supports(self, coverage, data_items):
        return len(data_items) == 1

    def connect(self, coverage, data_items, layer, options):
        data = data_items[0].path

        if coverage.grid.is_referenceable:
            vrt_path = join("/vsimem", uuid4().hex)
            reftools.create_rectified_vrt(data, vrt_path)
            data = vrt_path
            layer.setMetaData("eoxs_ref_data", data)

        if not layer.metadata.get("eoxs_wrap_dateline") == "true":
            layer.data = data
        else:
            sr = coverage.grid.spatial_reference
            extent = coverage.extent
            e = wrap_extent_around_dateline(extent, sr.srid)

            vrt_path = join("/vsimem", uuid4().hex)
            ds = gdal.Open(data)
            vrt_ds = create_simple_vrt(ds, vrt_path)
            size_x = ds.RasterXSize
            size_y = ds.RasterYSize

            dx = abs(e[0] - e[2]) / size_x
            dy = abs(e[1] - e[3]) / size_y

            vrt_ds.SetGeoTransform([e[0], dx, 0, e[3], 0, -dy])
            vrt_ds = None

            layer.data = vrt_path

    def disconnect(self, coverage, data_items, layer, options):
        if layer.metadata.get("eoxs_wrap_dateline") == "true":
            vsi.remove(layer.data)

        vrt_path = layer.metadata.get("eoxs_ref_data")
        if vrt_path:
            vsi.remove(vrt_path)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from eoxserver.core.util.timetools import parse_iso8601
from django.contrib.gis.geos import Polygon, MultiPolygon

from eoxserver.core.decoders import xml
from eoxserver.core.util.xmltools import parse


class DimapGeneralFormatReader(object):
    def test(self, obj):
        tree = parse(obj)
        return tree is not None and tree.getroot().tag == "Dimap_Document"

    def get_format_name(self, obj):
        return "dimap"

    def read(self, obj):
        tree = parse(obj)
        if self.test(tree):
            version = tree.xpath(
                "Metadadata_Id/METADATA_FORMAT/"
                "|Metadata_Identification/METADATA_FORMAT"
            )[0].get("version", "1.1")
            if version.startswith("1.1"):
                decoder = DimapGeneralFormat11Decoder(tree)
            elif version.startswith("2.0"):
                decoder = DimapGeneralFormat20Decoder(tree)
            else:
                raise Exception(
                    "DIMAP version '%s' is not supported." % version
                )

            values = {
                "identifier": decoder.identifier,
                "format": "dimap"
            }

            # in Dimap, pretty much everything is optional
            def cond_set(dct, key, value):
                if value is not None:
                    dct[key] = value

            # decode
            begin_time = decoder.begin_time
            end_time = decoder.end_time
            footprint = decoder.footprint
            projection = decoder.projection
            size = decoder.size
            gt = decoder.geotransform

            # calculate extent
            extent = None
            if size and gt:
                extent = (
                    gt[0], gt[3], gt[0] + gt[1] * size[0], gt[3] + gt[5] * size[1]
                )

            # set values
            cond_set(values, "begin_time", begin_time)
            cond_set(values, "end_time", begin_time)
            cond_set(values, "footprint", begin_time)
            cond_set(values, "size", begin_time)
            cond_set(values, "extent", begin_time)
            cond_set(values, "projection", begin_time)

            return values

        raise Exception("Could not parse from obj '%s'." % repr(obj))


def parse_date_or_datetime_11(string):
    value = parse_iso8601(string)
    if not value:
        raise Exception("Could not parse date or datetime from '%s'." % string)
    return value


def parse_footprint_11(elem):
    points = []

    for vertex in elem.findall("Vertex"):
        points.append((
            float(vertex.findtext("FRAME_LON")),
            float(vertex.findtext("FRAME_LAT"))
        ))

    if points[0] != points[-1]:
        points.append(points[0])

    return MultiPolygon(Polygon(points))


def parse_size_11(elem):
    return (int(elem.findtext("NCOLS")), int(elem.findtext("NROWS")))


def parse_geotransform_11(elem):
    values = (
        elem.findtext("ULXMAP"), elem.findtext("XDIM"), 0,
        elem.findtext("ULYMAP"), 0, elem.findtext("YDIM")
    )
    return map(float, values)


def parse_projection_11(elem):
    pass
    # TODO


class DimapGeneralFormat11Decoder(xml.Decoder):
    identifier  = xml.Parameter("Dataset_Id/DATASET_NAME/text()")
    begin_time  = xml.Parameter("Production/DATASET_PRODUCTION_DATE/text()", type=parse_date_or_datetime_11, num="?")
    end_time    = xml.Parameter("Production/DATASET_PRODUCTION_DATE/text()", type=parse_date_or_datetime_11, num="?")
    footprint   = xml.Parameter("Dataset_Frame", type=parse_footprint_11, num="?")
    size        = xml.Parameter("Raster_Dimensions", type=parse_size_11, num="?")
    geotransform = xml.Parameter("Geoposition/Geoposition_Insert", type=parse_geotransform_11, num="?")
    projection  = xml.Parameter("Coordinate_Reference_System/Horizontal_CS", type=parse_projection_11, num="?") # TODO


class DimapGeneralFormat20Decoder(xml.Decoder):
    identifier  = xml.Parameter("Dataset_Identification/DATASET_ID/text()")
    # TODO implement further parameters



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2015 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from django.db.models import Q

from eoxserver.core.decoders import kvp, enum
from eoxserver.core.util.xmltools import NameSpace
from eoxserver.core.util.timetools import parse_iso8601


class TimeExtension(object):
    """ Implementation of the OpenSearch `'Time' extension
    <http://www.opensearch.org/Specifications/OpenSearch/Extensions/Time/1.0/Draft_1>`_.
    """

    namespace = NameSpace(
        "http://a9.com/-/opensearch/extensions/time/1.0/", "time"
    )

    def filter(self, qs, parameters):
        decoder = TimeExtensionDecoder(parameters)

        start = decoder.start
        end = decoder.end
        relation = decoder.relation

        if start and end:
            if relation == "intersects":
                qs = qs.filter(Q(begin_time__lte=end) & Q(end_time__gte=start))
            elif relation == "contains":
                qs = qs.filter(Q(begin_time__lte=start) & Q(end_time__gte=end))
            elif relation == "during":
                qs = qs.filter(Q(begin_time__gte=start) & Q(end_time__lte=end))
            elif relation == "disjoint":
                qs = qs.filter(Q(begin_time__gt=end) | Q(end_time__lt=start))
            elif relation == "equals":
                qs = qs.filter(Q(begin_time=start) & Q(end_time=end))
        elif start:
            if relation == "intersects":
                qs = qs.filter(end_time__gte=start)
            elif relation == "contains":
                # not possible for a coverage to contain an open interval
                pass
            elif relation == "during":
                qs = qs.filter(begin_time__gte=start)
            elif relation == "disjoint":
                qs = qs.filter(end_time__lt=start)
            elif relation == "equals":
                qs = qs.filter(begin_time=start)
        elif end:
            if relation == "intersects":
                qs = qs.filter(begin_time__lte=end)
            elif relation == "contains":
                # see above
                pass
            elif relation == "during":
                qs = qs.filter(end_time__lte=end)
            elif relation == "disjoint":
                qs = qs.filter(begin_time__gt=end)
            elif relation == "equals":
                qs = qs.filter(end_time=end)
        return qs

    def get_schema(self, collection=None, model_class=None):
        return (
            dict(name="start", type="start"),
            dict(name="end", type="end"),
            dict(name="timerel", type="relation",
                options=["intersects", "contains", "disjoint", "equals"]
            )
        )


class TimeExtensionDecoder(kvp.Decoder):
    start = kvp.Parameter(num="?", type=parse_iso8601)
    end = kvp.Parameter(num="?", type=parse_iso8601)
    relation = kvp.Parameter(num="?",
        type=enum(("intersects", "contains", "disjoint", "equals"), False),
        default="intersects"
    )



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from eoxserver.core import Component, implements, UniqueExtensionPoint
from eoxserver.core.decoders import kvp, typelist, InvalidParameterException
from eoxserver.resources.coverages import crss
from eoxserver.services.subset import Subsets, Trim
from eoxserver.services.ows.interfaces import (
    ServiceHandlerInterface, GetServiceHandlerInterface
)
from eoxserver.services.ows.wms.util import (
    lookup_layers, parse_bbox, parse_time, int_or_str
)
from eoxserver.services.ows.wms.interfaces import WMSMapRendererInterface
from eoxserver.services.result import to_http_response
from eoxserver.services.ows.wms.exceptions import InvalidCRS


class WMS13GetMapHandler(object):
    methods = ['GET']

    service = ("WMS", None)
    versions = ("1.3.0", "1.3")
    request = "GetMap"

    def handle(self, request):
        decoder = WMS13GetMapDecoder(request.GET)

        bbox = decoder.bbox
        time = decoder.time
        crs = decoder.crs
        layers = decoder.layers

        if not layers:
            raise InvalidParameterException("No layers specified", "layers")

        srid = crss.parseEPSGCode(
            crs, (crss.fromShortCode, crss.fromURN, crss.fromURL)
        )
        if srid is None:
            raise InvalidCRS(crs, "crs")

        if crss.hasSwappedAxes(srid):
            miny, minx, maxy, maxx = bbox
        else:
            minx, miny, maxx, maxy = bbox

        subsets = Subsets((
            Trim("x", minx, maxx),
            Trim("y", miny, maxy),
        ), crs=crs)
        if time:
            subsets.append(time)

        # TODO: adjust way to get to renderer

        styles = decoder.styles

        if styles:
            styles = styles.split(',')

        from eoxserver.services.ows.wms.layerquery import LayerQuery

        render_map = LayerQuery().create_map(
            layers=layers, styles=styles, bbox=bbox, crs=crs,
            width=decoder.width, height=decoder.height,
            format=decoder.format, transparent=decoder.transparent,
            bgcolor=decoder.bgcolor,
            time=time,

            range=decoder.dim_range,

            bands=None,
            wavelengths=None,
            elevation=None,
            cql=decoder.cql,
        )


        from eoxserver.render.mapserver.map_renderer import MapserverMapRenderer

        return MapserverMapRenderer().render_map(render_map)

        # root_group = lookup_layers(layers, subsets, renderer.suffixes)

        # result, _ = renderer.render(
        #     root_group, request.GET.items(),
        #     width=int(decoder.width), height=int(decoder.height),
        #     time=decoder.time, bands=decoder.dim_bands, subsets=subsets,
        #     elevation=decoder.elevation,
        #     dimensions=dict(
        #         (key[4:], values) for key, values in decoder.dimensions
        #     )
        # )

        # return to_http_response(result)


def parse_transparent(value):
    value = value.upper()
    if value == 'TRUE':
        return True
    elif value == 'FALSE':
        return False
    raise ValueError("Invalid value for 'transparent' parameter.")


def parse_range(value):
    return map(float, value.split(','))


class WMS13GetMapDecoder(kvp.Decoder):
    layers = kvp.Parameter(type=typelist(str, ","), num=1)
    styles = kvp.Parameter(num="?")
    bbox   = kvp.Parameter(type=parse_bbox, num=1)
    time   = kvp.Parameter(type=parse_time, num="?")
    crs    = kvp.Parameter(num=1)
    width  = kvp.Parameter(num=1)
    height = kvp.Parameter(num=1)
    format = kvp.Parameter(num=1)
    bgcolor = kvp.Parameter(num='?')
    transparent = kvp.Parameter(num='?', default=False, type=parse_transparent)
    dim_bands = kvp.Parameter(type=typelist(int_or_str, ","), num="?")
    dim_range = kvp.Parameter(type=parse_range, num="?")
    elevation = kvp.Parameter(type=float, num="?")
    dimensions = kvp.MultiParameter(lambda s: s.startswith("dim_"), locator="dimension", num="*")

    cql = kvp.Parameter(num="?")



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from itertools import chain

from eoxserver.core import Component, env, implements, UniqueExtensionPoint
from eoxserver.core.decoders import kvp, typelist, InvalidParameterException
from eoxserver.resources.coverages import models, crss
from eoxserver.services.subset import Subsets, Trim, Slice
from eoxserver.services.ows.interfaces import (
    ServiceHandlerInterface, GetServiceHandlerInterface
)
from eoxserver.services.ows.wms.util import (
    lookup_layers, parse_bbox, parse_time, int_or_str
)
from eoxserver.services.ows.wms.interfaces import (
    WMSFeatureInfoRendererInterface
)
from eoxserver.services.result import to_http_response


class WMS11GetFeatureInfoHandler(Component):
    implements(ServiceHandlerInterface)
    implements(GetServiceHandlerInterface)

    renderer = UniqueExtensionPoint(WMSFeatureInfoRendererInterface)

    service = "WMS"
    versions = ("1.1", "1.1.0", "1.1.1")
    request = "GetFeatureInfo"

    def handle(self, request):
        decoder = WMS11GetFeatureInfoDecoder(request.GET)

        bbox = decoder.bbox
        time = decoder.time
        srs = decoder.srs
        layers = decoder.layers

        if not layers:
            raise InvalidParameterException("No layers specified", "layers")
    
        minx, miny, maxx, maxy = bbox

        subsets = Subsets((
            Trim("x", minx, maxx),
            Trim("y", miny, maxy),
        ), crs=srs)
        if time: 
            subsets.append(time)
        
        renderer = self.renderer
        root_group = lookup_layers(layers, subsets, renderer.suffixes)

        result, _ = renderer.render(
            root_group, request.GET.items(), request,
            time=decoder.time, bands=decoder.dim_bands
        )
        return to_http_response(result)


class WMS11GetFeatureInfoDecoder(kvp.Decoder):
    layers = kvp.Parameter(type=typelist(str, ","), num=1)
    styles = kvp.Parameter(num="?")
    bbox   = kvp.Parameter(type=parse_bbox, num=1)
    time   = kvp.Parameter(type=parse_time, num="?")
    srs    = kvp.Parameter(num=1)
    width  = kvp.Parameter(num=1)
    height = kvp.Parameter(num=1)
    format = kvp.Parameter(num=1)
    dim_bands = kvp.Parameter(type=typelist(int_or_str, ","), num="?")



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from lxml.builder import ElementMaker

from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.util.xmltools import NameSpace, NameSpaceMap
from eoxserver.services.ows.common.config import CapabilitiesConfigReader
from eoxserver.services.ows.common.v20.encoders import OWS20Encoder
from eoxserver.services.ows.common.v20.encoders import ns_xlink, ns_ows


ns_dseo = NameSpace("http://www.opengis.net/dseo/1.0", "dseo")
nsmap = NameSpaceMap(
    ns_xlink, ns_ows, ns_dseo
)

DSEO = ElementMaker(namespace=ns_dseo.uri, nsmap=nsmap)


class DSEO10CapabilitiesXMLEncoder(OWS20Encoder):
    def encode_capabilities(self, request, sections):
        conf = CapabilitiesConfigReader(get_eoxserver_config())

        all_sections = "all" in sections
        caps = []
        if all_sections or "serviceidentification" in sections:
            caps.append(self.encode_service_identification(
                "DSEO", conf, []
            ))

        if all_sections or "serviceprovider" in sections:
            caps.append(self.encode_service_provider(conf))

        if all_sections or "operationsmetadata" in sections:
            caps.append(self.encode_operations_metadata(
                request, "DSEO", ["1.0.0"]
            ))

        return DSEO(
            "Capabilities", *caps,
            version="1.0.0", updateSequence=conf.update_sequence
        )



def tag(*tags):
    """Decorator to add tags to a test class or method."""
    def decorator(obj):
        if len(getattr(obj, '__bases__', [])) > 1:
            setattr(obj, 'tags', set(tags).union(*[
                set(base.tags)
                for base in obj.__bases__
                if hasattr(base, 'tags')
            ]))
        elif hasattr(obj, 'tags'):
            obj.tags = obj.tags.union(tags)
        else:
            setattr(obj, 'tags', set(tags))
        return obj
    return decorator



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

""" This module provides base functionality for any other decoder class.
"""


from eoxserver.core.decoders import (
    ZERO_OR_ONE, ONE_OR_MORE, ANY, SINGLE_VALUES, WrongMultiplicityException,
    InvalidParameterException, MissingParameterException,
    MissingParameterMultipleException
)


class BaseParameter(property):
    """ Abstract base class for XML, KVP or any other kind of parameter.
    """

    def __init__(self, type=None, num=1, default=None):
        super(BaseParameter, self).__init__(self.fget)
        self.type = type or str
        self.num = num
        self.default = default

    def select(self, decoder):
        """ Interface method.
        """
        raise NotImplementedError

    @property
    def locator(self):
        return ""

    def fget(self, decoder):
        """ Property getter function.
        """

        results = self.select(decoder)
        count = len(results)

        locator = self.locator
        multiple = self.num not in SINGLE_VALUES

        # check the correct count of the result
        if not multiple and count > 1:
            raise WrongMultiplicityException(locator, "at most one", count)

        elif self.num == 1 and count == 0:
            raise MissingParameterException(locator)

        elif self.num == ONE_OR_MORE and count == 0:
            raise MissingParameterMultipleException(locator)

        elif isinstance(self.num, int) and count != self.num:
            raise WrongMultiplicityException(locator, self.num, count)

        # parse the value/values, or return the defaults
        if multiple:
            if count == 0 and self.num == ANY and self.default is not None:
                return self.default

            try:
                return [self.type(v) for v in results]
            except Exception as e:
                # let some more sophisticated exceptions pass
                if hasattr(e, "locator") or hasattr(e, "code"):
                    raise
                raise InvalidParameterException(str(e), locator)

        elif self.num == ZERO_OR_ONE and count == 0:
            return self.default

        elif self.type:
            try:
                return self.type(results[0])
            except Exception as e:
                # let some more sophisticated exceptions pass
                if hasattr(e, "locator") or hasattr(e, "code"):
                    raise
                raise InvalidParameterException(str(e), locator)

        return results[0]



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

""" This module contains facilities to help decoding KVP strings.
"""

from django.http import QueryDict
from django.utils.six import string_types, add_metaclass
from django.utils.six.moves.urllib.parse import parse_qs

from eoxserver.core.decoders.base import BaseParameter


class Parameter(BaseParameter):
    """ Parameter for KVP values.

        :param key: the lookup key; defaults to the property name of the
                    :class:`Decoder`
        :param type: the type to parse the raw value; by default the raw
                     string is returned
        :param num: defines how many times the key can be present; use any
                    numeric value to set it to a fixed count, "*" for any
                    number, "?" for zero or one time or "+" for one or more
                    times
        :param default: the default value
        :param locator: override the locator in case of exceptions
    """

    key = None

    def __init__(self, key=None, type=None, num=1, default=None, locator=None):
        super(Parameter, self).__init__(type, num, default)
        self.key = key.lower() if key is not None else None
        self._locator = locator

    def select(self, decoder, decoder_class=None):
        return [
            value for value in decoder._query_dict.get(self.key, [])
            if value != ""
        ]

    @property
    def locator(self):
        return self._locator or self.key


class MultiParameter(Parameter):
    """ Class for selecting different KVP parameters at once.

        :param selector: a function to determine if a key is used for the multi
                         parameter selection
        :param num: defines how many times the key can be present; use any
                    numeric value to set it to a fixed count, "*" for any
                    number, "?" for zero or one time or "+" for one or more
                    times
        :param default: the default value
        :param locator: override the locator in case of exceptions
    """

    def __init__(self, selector, num=1, default=None, locator=None):
        super(MultiParameter, self).__init__(
            "", lambda s: s, num, default, locator
        )
        self.key = selector

    def select(self, decoder):
        result = []
        for key, values in decoder._query_dict.items():
            if self.key(key):
                result.append((key, values))

        return result


class DecoderMetaclass(type):
    """ Metaclass for KVP Decoders to allow easy parameter declaration.
    """
    def __init__(cls, name, bases, dct):
        # set the "key" attribute of any parameter to the parameters name
        # if no other key was specified.
        for key, value in dct.items():
            if isinstance(value, Parameter) and value.key is None:
                value.key = key.lower()

        super(DecoderMetaclass, cls).__init__(name, bases, dct)


@add_metaclass(DecoderMetaclass)
class Decoder(object):
    """ Base class for KVP decoders.

    :param params: an instance of either :class:`dict`,
                   :class:`django.http.QueryDict` or :class:`basestring` (which
                   will be parsed using :func:`cgi.parse_qs`)

    Decoders should be used as such:
    ::

        from eoxserver.core.decoders import kvp
        from eoxserver.core.decoders import typelist

        class ExampleDecoder(kvp.Decoder):
            mandatory_param = kvp.Parameter(num=1)
            list_param = kvp.Parameter(type=typelist(separator=","))
            multiple_param = kvp.Parameter("multi", num="+")
            optional_param = kvp.Parameter(num="?", default="default_value")

        decoder = ExampleDecoder(
            "mandatory_param=value"
            "&list_param=a,b,c"
            "&multi=a&multi=b&multi=c"
        )

        print decoder.mandatory_param
        print decoder.list_param
        print decoder.multiple_param
        print decoder.optional_param
    """

    __metaclass__ = DecoderMetaclass

    def __init__(self, params):
        query_dict = {}
        if isinstance(params, QueryDict):
            for key, values in params.lists():
                query_dict[key.lower()] = values

        elif isinstance(params, string_types):
            tmp = parse_qs(params)
            for key, values in tmp.items():
                query_dict[key.lower()] = values

        elif isinstance(params, dict):
            for key, value in params.items():
                query_dict[key.lower()] = (value,)

        else:
            raise ValueError(
                "Decoder input '%s' not supported." % type(params).__name__
            )
        self.kvp = params
        self._query_dict = query_dict



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.core import Component, implements, UniqueExtensionPoint
from eoxserver.core.decoders import kvp, typelist, InvalidParameterException
from eoxserver.resources.coverages import crss
from eoxserver.services.subset import Subsets, Trim
from eoxserver.services.ows.interfaces import (
    ServiceHandlerInterface, GetServiceHandlerInterface
)
from eoxserver.services.ows.wms.util import (
    lookup_layers, parse_bbox
)
from eoxserver.services.ows.wms.interfaces import WMSMapRendererInterface
from eoxserver.services.result import to_http_response
from eoxserver.services.ows.wms.exceptions import InvalidCRS
from eoxserver.services.ows.wms.v10.decoders import WMS10GetMapDecoder


class WMS10GetMapHandler(object):
    service = ("WMS", None)
    versions = ("1.0", "1.0.0")
    request = "GetMap"

    def handle(self, request):
        decoder = WMS10GetMapDecoder(request.GET)

        bbox = decoder.bbox
        srs = decoder.srs
        layers = decoder.layers

        if not layers:
            raise InvalidParameterException("No layers specified", "layers")

        srid = crss.parseEPSGCode(
            srs, (crss.fromShortCode, crss.fromURN, crss.fromURL)
        )
        if srid is None:
            raise InvalidCRS(srs, "srs")

        # WMS 1.1 knows no swapped axes
        minx, miny, maxx, maxy = bbox

        subsets = Subsets((
            Trim("x", minx, maxx),
            Trim("y", miny, maxy),
        ), crs=srs)

        root_group = lookup_layers(layers, subsets)

        result, _ = self.renderer.render(
            root_group, request.GET.items(), subsets=subsets,
            width=int(decoder.width), height=int(decoder.height)
        )
        return to_http_response(result)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import logging
from datetime import datetime

from django.utils.six.moves.urllib.parse import unquote
from django.utils.six import binary_type, b
from lxml import etree

from eoxserver.contrib import mapserver as ms
from eoxserver.contrib import gdal, vsi
from eoxserver.render.coverage import objects
from eoxserver.resources.coverages import models, crss
from eoxserver.resources.coverages.formats import getFormatRegistry
from eoxserver.services.exceptions import NoSuchCoverageException
from eoxserver.services.ows.wcs.v20.encoders import WCS20EOXMLEncoder
from eoxserver.services.ows.wcs.v20.util import (
    ScaleSize, ScaleExtent, ScaleAxis
)
from eoxserver.services.mapserver.wcs.base_renderer import (
    BaseRenderer, is_format_supported
)
from eoxserver.services.ows.version import Version
from eoxserver.services.result import result_set_from_raw_data, ResultBuffer
from eoxserver.services.exceptions import (
    RenderException, OperationNotSupportedException,
    InterpolationMethodNotSupportedException, InvalidOutputCrsException
)
from eoxserver.services.mapserver.connectors import get_connector_by_test


logger = logging.getLogger(__name__)

INTERPOLATION_TRANS = {
    "nearest-neighbour": "NEAREST",
    "linear": "BILINEAR",
    "bilinear": "BILINEAR",
    "cubic": "BICUBIC",
    "average": "AVERAGE"
}


class RectifiedCoverageMapServerRenderer(BaseRenderer):
    """ A coverage renderer for rectified coverages. Uses mapserver to process
        the request.
    """

    # ReferenceableDatasets are not handled in WCS >= 2.0
    versions_full = (Version(1, 1), Version(1, 0))
    versions_partly = (Version(2, 0),)
    versions = versions_full + versions_partly

    # handles_full = (
    #     models.RectifiedDataset,
    #     models.RectifiedStitchedMosaic,
    #     models.ReferenceableDataset
    # )

    # handles_partly = (models.RectifiedDataset, models.RectifiedStitchedMosaic)
    # handles = handles_full + handles_partly

    # connectors = ExtensionPoint(ConnectorInterface)
    # layer_factories = ExtensionPoint(LayerFactoryInterface)

    def supports(self, params):
        # return (
        #     (
        #         params.version in self.versions_full and
        #     and issubclass(params.coverage.real_type, self.handles_full))
        #     or
        #     (params.version in self.versions_partly
        #     and issubclass(params.coverage.real_type, self.handles_partly))
        # )
        return params.version in self.versions and not params.coverage.grid.is_referenceable

    def render(self, params):
        # get coverage related stuff
        coverage = params.coverage

        # ReferenceableDataset are not supported in WCS < 2.0
        if params.coverage.grid.is_referenceable and params.version:
            raise NoSuchCoverageException((coverage.identifier,))

        data_locations = self.arraydata_locations_for_coverage(coverage)

        range_type = coverage.range_type
        bands = list(range_type)

        subsets = params.subsets

        if subsets:
            subsets.srid  # this automatically checks the validity

        # create and configure map object
        map_ = self.create_map()

        env = {}
        for data_location in data_locations:
            env.update(data_location.env)
        gdal.set_env(env, False)

        # configure outputformat
        native_format = self.get_native_format(coverage, data_locations)
        if native_format and get_format_by_mime(native_format) is None:
            native_format = "image/tiff"

        frmt = params.format or native_format

        if frmt is None:
            raise RenderException("Format could not be determined", "format")

        mime_type, frmt = split_format(frmt)

        imagemode = ms.gdalconst_to_imagemode(bands[0].data_type)
        time_stamp = datetime.now().strftime("%Y%m%d%H%M%S")
        basename = "%s_%s" % (coverage.identifier, time_stamp)
        of = create_outputformat(
            mime_type, frmt, imagemode, basename,
            getattr(params, "encoding_params", {})
        )

        map_.appendOutputFormat(of)
        map_.setOutputFormat(of)

        # TODO: use layer factory here
        layer = self.layer_for_coverage(coverage, native_format, params.version)

        map_.insertLayer(layer)
        connector = get_connector_by_test(coverage, data_locations)

        if not connector:
            raise OperationNotSupportedException(
                "Could not find applicable layer connector.", "coverage"
            )

        try:
            connector.connect(coverage, data_locations, layer, {})
            # create request object and dispatch it against the map
            request = ms.create_request(
                self.translate_params(params, range_type)
            )
            request.setParameter("format", mime_type)
            raw_result = ms.dispatch(map_, request)

        finally:
            # perform any required layer related cleanup
            connector.disconnect(coverage, data_locations, layer, {})

        result_set = result_set_from_raw_data(raw_result)

        if params.version == Version(2, 0):
            mediatype = getattr(params, "mediatype", None)
            if mediatype in ("multipart/mixed", "multipart/related"):
                with vsi.TemporaryVSIFile.from_buffer(result_set[1].data) as f:
                    ds = gdal.Open(f.name)
                    grid = objects.Grid.from_gdal_dataset(ds)

                    # get the output CRS definition
                    crs = params.outputcrs or subsets.crs or 'imageCRS'
                    if crs == 'imageCRS':
                        crs = coverage.grid.coordinate_reference_system
                    grid._coordinate_reference_system = crs

                    origin = objects.Origin.from_gdal_dataset(ds)
                    size = [ds.RasterXSize, ds.RasterYSize]

                range_type = coverage.range_type
                if params.rangesubset:
                    range_type = range_type.subset(params.rangesubset)

                coverage._grid = grid
                coverage._origin = origin
                coverage._size = size
                coverage._range_type = range_type
                if isinstance(result_set[1].filename, binary_type):
                    file_name = result_set[1].filename.decode()
                else:
                    file_name = result_set[1].filename

                reference = 'cid:coverage/%s' % file_name

                encoder = WCS20EOXMLEncoder()

                if not isinstance(coverage, objects.Mosaic):
                    tree = encoder.encode_rectified_dataset(
                        coverage,
                        getattr(params, "http_request", None),
                        reference,
                        mime_type,
                        subsets.bounding_polygon(coverage) if subsets else None
                    )
                else:
                    tree = encoder.encode_rectified_stitched_mosaic(
                        coverage,
                        getattr(params, "http_request", None),
                        reference,
                        mime_type,
                        subsets.bounding_polygon(coverage) if subsets else None
                    )

                result_set[0] = ResultBuffer(
                    encoder.serialize(tree),
                    encoder.content_type
                )

        # "default" response
        return result_set

    def translate_params(self, params, range_type):
        """ "Translate" parameters to be understandable by mapserver.
        """

        if params.version.startswith("2.0"):
            for key, value in params:
                if key == 'coverageid':
                    try:
                        models.identifier_validators[0](value)
                    except:
                        value = 'not-ncname'
                    yield key, value

                elif key == "interpolation":
                    interpolation = INTERPOLATION_TRANS.get(value)
                    if not interpolation:
                        raise InterpolationMethodNotSupportedException(
                            "Interpolation method '%s' is not supported."
                            % value
                        )
                    yield key, value

                else:
                    yield key, value

            rangesubset = params.rangesubset
            if rangesubset:
                yield "rangesubset", ",".join(
                    map(str, rangesubset.get_band_indices(range_type, 1))
                )

            # TODO: this only works in newer MapServer implementations
            # (since 6.4?).
            SCALE_AVAILABLE = ms.msGetVersionInt() > 60401
            scalefactor = params.scalefactor
            if scalefactor is not None:
                if SCALE_AVAILABLE:
                    yield "scalefactor", str(scalefactor)
                else:
                    raise RenderException(
                        "'ScaleFactor' is not supported by MapServer in the "
                        "current version.", "scalefactor"
                    )

            for scale in params.scales:
                scaleaxes = []
                if isinstance(scale, ScaleSize):
                    yield "size", "%s(%d)" % (scale.axis, scale.size)
                elif isinstance(scale, ScaleExtent):
                    yield "size", "%s(%d)" % (scale.axis, scale.high-scale.low)
                elif isinstance(scale, ScaleAxis):
                    if SCALE_AVAILABLE:
                        scaleaxes.append(scale)
                    else:
                        raise RenderException(
                            "'ScaleAxes' is not supported by MapServer in the "
                            "current version.", "scaleaxes"
                        )

                if scaleaxes:
                    yield "scaleaxes", ",".join(
                        "%s(%f)" % (scale.axis, scale.scale)
                        for scale in scaleaxes
                    )

            if params.outputcrs is not None:
                srid = crss.parseEPSGCode(params.outputcrs,
                    (crss.fromURL, crss.fromURN, crss.fromShortCode)
                )
                if srid is None:
                    raise InvalidOutputCrsException(
                        "Failed to extract an EPSG code from the OutputCRS URI "
                        "'%s'." % params.outputcrs
                    )
                yield "outputcrs", params.outputcrs

        else:
            for key, value in params:
                yield key, value


def split_format(frmt):
    parts = unquote(frmt).split(";")
    mime_type = parts[0]
    options = map(
        lambda kv: map(lambda i: i.strip(), kv.split("=")), parts[1:]
    )
    return mime_type, options


def create_outputformat(mime_type, options, imagemode, basename, parameters):
    """ Returns a ``mapscript.outputFormatObj`` for the given format name and
        imagemode.
    """

    reg_format = get_format_by_mime(mime_type)

    if not reg_format:
        raise RenderException(
            "Unsupported output format '%s'." % mime_type, "format"
        )

    outputformat = ms.outputFormatObj(reg_format.driver, "custom")
    outputformat.name = reg_format.wcs10name
    outputformat.mimetype = reg_format.mimeType
    outputformat.extension = reg_format.defaultExt
    outputformat.imagemode = imagemode

    # for key, value in options:
    #    outputformat.setOption(str(key), str(value))

    if mime_type == "image/tiff":
        _apply_gtiff(outputformat, **parameters)

    filename = basename + reg_format.defaultExt
    outputformat.setOption("FILENAME", str(filename))

    return outputformat


def _apply_gtiff(outputformat, compression=None, jpeg_quality=None,
                 predictor=None, interleave=None, tiling=False,
                 tilewidth=None, tileheight=None):

    logger.info("Applying GeoTIFF parameters.")

    if compression:
        if compression.lower() == "huffman":
            compression = "CCITTRLE"
        outputformat.setOption("COMPRESS", str(compression.upper()))

    if jpeg_quality is not None:
        outputformat.setOption("JPEG_QUALITY", str(jpeg_quality))

    if predictor:
        pr = ["NONE", "HORIZONTAL", "FLOATINGPOINT"].index(predictor.upper())
        if pr == -1:
            raise ValueError("Invalid compression predictor '%s'." % predictor)
        outputformat.setOption("PREDICTOR", str(pr + 1))

    if interleave:
        outputformat.setOption("INTERLEAVE", str(interleave.upper()))

    if tiling:
        outputformat.setOption("TILED", "YES")
        if tilewidth is not None:
            outputformat.setOption("BLOCKXSIZE", str(tilewidth))
        if tileheight is not None:
            outputformat.setOption("BLOCKYSIZE", str(tileheight))


def get_format_by_mime(mime_type):
    """ Convenience function to return an enabled format descriptior for the
        given mime type or WCS 1.0 format name. Returns ``None``, if none
        applies.
    """

    registry = getFormatRegistry()
    reg_format = registry.getFormatByMIME(mime_type)

    if not reg_format:
        wcs10_frmts = registry.getFormatsByWCS10Name(mime_type)
        if wcs10_frmts:
            reg_format = wcs10_frmts[0]

    if reg_format and not is_format_supported(reg_format.mimeType):
        return None

    return reg_format



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import os.path
import shutil
import tarfile
import zipfile
import fnmatch
from django.utils.six.moves.urllib import parse, request

import ftplib
import glob

from django.conf import settings
from django.utils.module_loading import import_string

from eoxserver.contrib import vsi
from eoxserver.backends.config import DEFAULT_EOXS_STORAGE_HANDLERS


class BaseStorageHandler(object):
    """ Storage Handlers must conform to the context manager protocol
    """

    name = None  # short name of the storage handler

    allows_child_storages = False
    allows_parent_storage = False

    is_local = False

    def __enter__(self):
        """ Perform setup actions. Will be called before ``retrieve`` and
            ``list_files``.
        """
        return self

    def __exit__(self, type, value, traceback):
        """ Perform teardown actions. Will be called when the storage is no
            longer used.
        """
        pass

    def retrieve(self, location, path):
        """ Retrieve the file specified by `location` under the given local
            `path`. The path is only a hint, when a string is returned, this
            indicates that the file was instead stored in that location.
            Should be implemented by storage handlers that deal with
            files or similar objects.
        """
        raise NotImplementedError

    def list_files(self, glob_pattern=None):
        """ List the files in that storage, optionally filtered by a glob. Should
            be implemented for storages dealing with files, when possible.
        """
        raise NotImplementedError

    def get_vsi_path(self, location):
        """ Get the VSI file path for the file specified by location. This path
            can be used in GDAL based APIs to directly adress files.
        """
        raise NotImplementedError

    def get_vsi_env(self):
        """
        """
        return {}

    @classmethod
    def test(cls, locator):
        """ Check if a locator refers to a storage that can be handled by this
            handler class.
        """
        raise NotImplementedError


class ZIPStorageHandler(BaseStorageHandler):
    """Implementation of the storage interface for ZIP storages.
    """

    name = "ZIP"

    allows_child_storages = True
    allows_parent_storage = True

    is_local = True

    def __init__(self, package_filename):
        self.package_filename = package_filename
        self.zipfile = None

    def __enter__(self):
        self.zipfile = zipfile.ZipFile(self.package_filename, "r")
        return self

    def __exit__(self, type, value, traceback):
        self.zipfile.close()
        self.zipfile = None

    def retrieve(self, location, path):
        infile = self.zipfile.open(location)
        with open(path, "wb") as outfile:
            shutil.copyfileobj(infile, outfile)
        return True, path

    def list_files(self, glob_pattern=None):
        filenames = self.zipfile.namelist()
        if glob_pattern:
            filenames = fnmatch.filter(filenames, glob_pattern)
        return filenames

    def get_vsi_path(self, location):
        return '/vsizip/%s/%s' % (self.package_filename, location)

    @classmethod
    def test(cls, locator):
        return zipfile.is_zipfile(locator)


class TARStorageHandler(BaseStorageHandler):
    """Implementation of the storage interface for ZIP storages.
    """

    name = "TAR"

    allows_child_storages = True
    allows_parent_storage = True

    is_local = True

    def __init__(self, package_filename):
        self.package_filename = package_filename
        self.tarfile = None

    def __enter__(self):
        self.tarfile = tarfile.TarFile(self.package_filename, "r")
        return self

    def __exit__(self, type, value, traceback):
        self.tarfile.close()
        self.tarfile = None

    def retrieve(self, location, path):
        self.tarfile.extract(location, path)
        return True, path

    def list_files(self, glob_pattern=None):
        filenames = self.tarfile.getnames()
        if glob_pattern:
            filenames = fnmatch.filter(filenames, glob_pattern)
        return filenames

    def get_vsi_path(self, location):
        return '/vsitar/%s/%s' % (self.package_filename, location)

    @classmethod
    def test(cls, locator):
        try:
            return tarfile.is_tarfile(locator)
        except IOError:
            return False


class DirectoryStorageHandler(BaseStorageHandler):
    """
    """

    name = 'directory'

    allows_child_storages = True
    allows_parent_storage = True

    is_local = True

    def __init__(self, dirpath):
        self.dirpath = dirpath

    def retrieve(self, location, path):
        return False, os.path.join(self.dirpath, location)

    def list_files(self, glob_pattern=None):
        glob_pattern = glob_pattern or '*'
        return glob.glob(os.path.join(self.dirpath, glob_pattern))

    def get_vsi_path(self, location):
        return os.path.join(self.dirpath, location)

    @classmethod
    def test(cls, locator):
        return os.path.isdir(locator)


class HTTPStorageHandler(BaseStorageHandler):
    """
    """

    name = 'HTTP'

    allows_child_storages = True
    allows_parent_storage = False

    def __init__(self, url):
        self.url = url

    def retrieve(self, location, path):
        request.urlretrieve(parse.urljoin(self.url, location), path)
        return True, path

    def get_vsi_path(self, location):
        return '/vsicurl/%s' % parse.urljoin(self.url, location)

    @classmethod
    def test(cls, locator):
        try:
            return urlparse(locator).scheme.lower() in ('http', 'https')
        except:
            return False


class FTPStorageHandler(BaseStorageHandler):
    """
    """

    name = 'FTP'

    allows_parent_storage = True
    allows_parent_storage = False

    def __init__(self, url):
        self.url = url
        self.parsed_url = urlparse(url)
        self.ftp = None

    def __enter__(self):
        self.ftp = ftplib.FTP()
        self.ftp.connect(self.parsed_url.hostname, self.parsed_url.port)
        self.ftp.login(self.parsed_url.username, self.parsed_url.password)
        return self

    def __exit__(self, type, value, traceback):
        self.ftp.quit()
        self.ftp = None

    def retrieve(self, location, path):
        cmd = "RETR %s" % os.path.join(self.parsed_url.path, location)
        with open(path, 'wb') as local_file:
            self.ftp.retrbinary(cmd, local_file.write)
        return True, path

    def list_files(self, location, glob_pattern=None):
        try:
            filenames = self.ftp.nlst(location)
        except ftplib.error_perm as resp:
            if str(resp).startswith("550"):
                filenames = []
            else:
                raise
        if glob_pattern:
            filenames = fnmatch.filter(filenames, glob_pattern)
        return filenames

    def get_vsi_path(self, location):
        return '/vsicurl/%s' % parse.urljoin(self.url, location)

    @classmethod
    def test(cls, locator):
        try:
            return urlparse(locator).scheme.lower() == 'ftp'
        except:
            return False


class SwiftStorageHandler(BaseStorageHandler):
    name = 'swift'

    allows_parent_storage = False
    allows_child_storages = True

    def __init__(self, url):
        self.container = url

    def retrieve(self, location, path):
        pass

    def list_files(self, location, glob_pattern=None):
        return []

    def get_vsi_path(self, location):
        return vsi.join('/vsiswift/%s' % self.container, location)

    @classmethod
    def test(cls, locator):
        return False


class S3StorageHandler(BaseStorageHandler):
    name = 'S3'

    allows_parent_storage = False
    allows_child_storages = True

    def __init__(self, url):
        self.bucket = url

    def retrieve(self, location, path):
        pass

    def list_files(self, location, glob_pattern=None):
        return []

    def get_vsi_path(self, location):
        import logging
        logger = logging.getLogger(__name__)

        # logger.debug()


        base_path = '/vsis3/%s' % self.bucket if self.bucket else '/vsis3'
        return vsi.join(base_path, location)

    @classmethod
    def test(cls, locator):
        return False


# API to setup and retrieve the configured storage handlers

STORAGE_HANDLERS = None


def _setup_storage_handlers():
    """ Setup the storage handlers. Uses the ``EOXS_STORAGE_HANDLERS`` setting
        which falls back to the ``DEFAULT_EOXS_STORAGE_HANDLERS``
    """
    global STORAGE_HANDLERS
    specifiers = getattr(
        settings, 'EOXS_STORAGE_HANDLERS', DEFAULT_EOXS_STORAGE_HANDLERS
    )
    STORAGE_HANDLERS = [import_string(specifier) for specifier in specifiers]


def get_handlers():
    if STORAGE_HANDLERS is None:
        _setup_storage_handlers()

    return STORAGE_HANDLERS


def get_handler_by_test(locator):
    """ Test the given locator with the configured storage handlers and return the stora
    """
    if STORAGE_HANDLERS is None:
        _setup_storage_handlers()

    for storage_handler_cls in STORAGE_HANDLERS:
        try:
            if storage_handler_cls.test(locator):
                return storage_handler_cls(locator)
        except AttributeError:
            pass


def get_handler_class_by_name(name):
    if STORAGE_HANDLERS is None:
        _setup_storage_handlers()

    for storage_handler_cls in STORAGE_HANDLERS:
        try:
            if storage_handler_cls.name == name:
                return storage_handler_cls
        except AttributeError:
            pass


def get_handler_class_for_model(storage_model):
    return get_handler_class_by_name(storage_model.storage_type)


def get_handler_for_model(storage_model):
    return get_handler_class_for_model(storage_model)(storage_model.url)
